"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Leaf = exports.Tree = exports.Node = exports.traversePath = exports.splitPath = exports.nodeToJsValue = exports.treeFromJsObject = exports.treeFromEntries = exports.nodeFromEntry = void 0;
function nodeFromEntry(entry, path) {
    switch (entry[0]) {
        case 'leaf': return new Leaf([...path, entry[1]], entry[2]);
        case 'tree': return treeFromEntries(entry[2], [...path, entry[1]]);
        default: throw new Error(`Unrecognized entry type '${entry[0]}'`);
    }
}
exports.nodeFromEntry = nodeFromEntry;
function treeFromEntries(entries, path = []) {
    const tree = new Tree(path);
    for (const e of entries)
        tree.addNode(nodeFromEntry(e, tree.path));
    return tree;
}
exports.treeFromEntries = treeFromEntries;
function treeFromJsObject(obj, path = []) {
    const tree = new Tree(path);
    for (const key in obj) {
        const value = obj[key];
        if (typeof value === 'object') {
            tree.addNode(treeFromJsObject(value, [...path, key]));
        }
        else {
            tree.addNode(new Leaf([...path, key], value));
        }
    }
    return tree;
}
exports.treeFromJsObject = treeFromJsObject;
function nodeToJsValue(node) {
    if (node.isLeaf)
        return node.value;
    const result = {};
    for (const [k, v] of node.entries) {
        result[k] = nodeToJsValue(v);
    }
    return result;
}
exports.nodeToJsValue = nodeToJsValue;
function splitPath(path) {
    const { length } = path;
    if (!length)
        throw new Error('Expected path with at least 1 entry');
    return [path.slice(0, length - 1), path[length - 1]];
}
exports.splitPath = splitPath;
function traversePath(tree, path, wrongNode) {
    let current = tree;
    for (let i = 0; i < path.length; i++) {
        const currentPath = path.slice(0, i + 1);
        if (current.isLeaf) {
            current = wrongNode === null || wrongNode === void 0 ? void 0 : wrongNode(currentPath, current);
        }
        if (!(current instanceof Tree))
            return undefined;
        current = current.getNode(path[i]) || (wrongNode === null || wrongNode === void 0 ? void 0 : wrongNode(currentPath));
        if (!current)
            return undefined;
    }
    return current;
}
exports.traversePath = traversePath;
class Node {
    constructor(isLeaf, path) {
        this.isLeaf = isLeaf;
        this.path = path;
    }
}
exports.Node = Node;
function isParentOf(parent, node) {
    const { path: parentPath } = parent;
    const { path: nodePath } = node;
    if (nodePath.length !== parentPath.length + 1)
        return false;
    for (let i = 0; i < parentPath.length; i++) {
        if (nodePath[i] !== parentPath[i])
            return false;
    }
    return true;
}
function expectParentOf(parent, node) {
    if (isParentOf(parent, node))
        return;
    throw new Error(`Expected tree '${parent.path}' to be parent of '${node.path}'`);
}
class Tree extends Node {
    constructor(path = [], entries) {
        super(false, path);
        this.entries = new Map([...(entries || [])]);
    }
    addNode(node) {
        expectParentOf(this, node);
        this.entries.set(node.path[this.path.length], node);
    }
    getNode(key) {
        return this.entries.get(key);
    }
    removeNode(nodeOrKey) {
        if (nodeOrKey instanceof Node) {
            expectParentOf(this, nodeOrKey);
            this.entries.delete(nodeOrKey.path[nodeOrKey.path.length - 1]);
        }
        else {
            this.entries.delete(nodeOrKey);
        }
    }
    toSerializedEntry() {
        const entries = [];
        for (const [key, node] of this.entries)
            entries.push(node.toSerializedEntry());
        entries.sort((a, b) => a[1] < b[1] ? -1 : 1);
        return ['tree', this.path[this.path.length - 1], entries];
    }
    clone() {
        const result = new Tree(this.path);
        for (const e of this.entries)
            result.addNode(e[1].clone());
        return result;
    }
}
exports.Tree = Tree;
class Leaf extends Node {
    constructor(path, value) {
        super(true, path);
        this.value = value;
    }
    toSerializedEntry() {
        return ['leaf', this.path[this.path.length - 1], this.value];
    }
    clone() {
        return this; // Immutable anyway
    }
}
exports.Leaf = Leaf;
//# sourceMappingURL=tree.js.map